import logging
from collections import defaultdict
from datetime import datetime, timedelta
from enum import Enum

from aio_pika import connect_robust

from kontiki.delegate import ServiceDelegate
from kontiki.messaging.common import create_tls_context, get_amqp_url
from kontiki.messaging.serialization import Serializer
from kontiki.registry.common import declare_registry_admin_exchange
from kontiki.registry.server.delegates.event_tracking import EventTracker
from kontiki.registry.server.delegates.exception_tracking import ExceptionTracker
from kontiki.registry.server.delegates.heartbeat_manager import HeartbeatManager
from kontiki.registry.server.delegates.registry import Registry

# -----------------------------------------------------------------------------


def make_serializable(obj):
    if isinstance(obj, dict):
        return {k: make_serializable(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [make_serializable(item) for item in obj]
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, datetime):
        return obj.isoformat()
    if hasattr(obj, "__dict__"):
        return make_serializable(obj.__dict__)
    if isinstance(obj, bytes):
        return obj.decode("utf-8", errors="replace")
    return obj


# -----------------------------------------------------------------------------


class ServiceStatus(Enum):
    ACTIVE = "active"
    DEGRADED = "degraded"
    DOWN = "down"


class ServiceRegistryCore(ServiceDelegate):
    def __init__(self):
        self.connection = None
        self.channel = None
        self.registry_admin_exchange = None

        self.heartbeats = {}
        self.default_timeout_factor = 3

        self.event_tracker = EventTracker(self)
        self.registry = Registry(self)
        self.heartbeat_manager = HeartbeatManager(self)
        self.exception_tracker = ExceptionTracker(self)
        super().__init__()

    async def setup(self):
        logging.debug("ServiceRegistration Setup")
        amqp_url = get_amqp_url(self.container.config)

        ssl_ctx = create_tls_context(self.container.config)
        self.connection = await connect_robust(amqp_url, ssl_context=ssl_ctx)

        self.channel = await self.connection.channel()
        await self.channel.set_qos(prefetch_count=10)
        self.registry_admin_exchange = await declare_registry_admin_exchange(
            self.channel
        )

        # Sets serializer
        self.serializer = Serializer(self.container.config)

        # Sets delegates up.
        await self.event_tracker.setup()
        await self.registry.setup()
        await self.heartbeat_manager.setup()
        await self.exception_tracker.setup()

    async def stop(self):
        if self.connection:
            await self.connection.close()

    async def create_and_consume_queue(self, routing_key, callback):
        queue_name = f"{routing_key}.queue"
        try:
            queue = await self.channel.declare_queue(queue_name, durable=True)
            await queue.bind(self.registry_admin_exchange, routing_key=routing_key)
            logging.debug("Binding %s with rkey %s", queue_name, routing_key)
            await queue.consume(callback)
            logging.debug("Created and bound queue: %s", queue_name)
        except Exception as e:
            logging.error("Error creating or consuming queue %s: %s", queue_name, e)

    def _get_instance_status(self, instance_id, service_name, timeout):
        logging.debug("Get %s-%s status.", service_name, instance_id)
        last_heartbeat = self.heartbeat_manager.get_last_heartbeat(
            service_name, instance_id
        )
        is_degraded = self.heartbeat_manager.is_degraded(service_name, instance_id)

        is_late = datetime.now() - last_heartbeat > timedelta(seconds=timeout)
        if not last_heartbeat or is_late:
            return ServiceStatus.DOWN.value
        if is_degraded:
            return ServiceStatus.DEGRADED.value
        return ServiceStatus.ACTIVE.value

    def _get_timeout(self, heartbeat_interval):
        if heartbeat_interval is None:
            heartbeat_interval = 10
        timeout = heartbeat_interval * self.default_timeout_factor
        logging.debug("Calculated timeout = %s", timeout)
        return timeout

    def get_services(self, status=None):
        filtered_services = defaultdict(dict)
        try:
            for service_name, instances in self.registry.services.items():
                for instance_id, data in instances.items():
                    heartbeat_interval = data.get("heartbeat_interval", 10)
                    current_status = self._get_instance_status(
                        instance_id, service_name, self._get_timeout(heartbeat_interval)
                    )
                    if not status or current_status == status:
                        filtered_services[service_name][instance_id] = {
                            "status": current_status,
                            "metadata": data,
                        }
        except Exception as e:
            logging.error("Error while getting services: (%s)", e)

        return dict(filtered_services)

    def get_events(self):
        return make_serializable(self.event_tracker.events)

    def get_exceptions(self):
        return make_serializable(self.exception_tracker.exceptions)



import asyncio
import logging

from kontiki.container import ServiceContainer
from kontiki.messaging.publisher.messenger import Messenger

# -----------------------------------------------------------------------------

DEFAULT_CONF = {"kontiki": {"amqp": {"url": "amqp://guest:guest@localhost/"}}}

# -----------------------------------------------------------------------------


class MockServiceManager:
    def __init__(self, log_file, log_level="INFO", messenger=None, default_config=None):
        self.mock_args = {}
        self.mocks = {}
        self.event_conditions = {}
        self.messenger = messenger if messenger else Messenger(standalone=True)
        self._setup = False
        self.default_dummy_added = False
        self.default_service_config = default_config or DEFAULT_CONF

        format_ = (
            "%(asctime)s - %(levelname)s - %(message)s - %(filename)s - %(lineno)d"
        )

        self.logging_conf = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {"default": {"format": format_}},
            "handlers": {
                "file": {
                    "class": "logging.FileHandler",
                    "formatter": "default",
                    "level": log_level,
                    "filename": log_file,
                }
            },
            "root": {"level": "DEBUG", "handlers": ["file"]},
        }

    # Service management.
    def add(self, mock_class, config):
        name = mock_class.name if hasattr(mock_class, "name") else mock_class.__name__
        self.mock_args[name] = (mock_class, config)

    async def start(self):
        if not self.mock_args:
            # We add a DummyService so that the tester can use the MockServiceRunner
            # just to publish events or call a remote function without writing a
            # MockService they don't really need.
            class DefaultDummyService:
                name = "dummy_service"

                async def do_nothing(self):
                    pass

            logging.info("No services registered. Adding DefaultDummyService.")
            dummy_config = {"logging": self.logging_conf} | self.default_service_config
            self.add(DefaultDummyService, config=dummy_config)
            self.default_dummy_added = True

        for mock_name, args in self.mock_args.items():
            logging.info("Starting %s mock service with args = %s", mock_name, args)

            self.mocks[mock_name] = await self.start_service(*args)
            self.event_conditions[mock_name] = asyncio.Condition()

    async def stop(self):
        for _, (container, _) in self.mocks.items():
            await container.stop()
        self.mocks.clear()
        self.event_conditions.clear()
        if self.default_dummy_added:
            logging.info("DefaultDummyService stopped.")

        if hasattr(self.messenger, "stop") and self.messenger:
            await self.messenger.stop()

    async def start_service(self, service_cls, config):
        config["logging"] = self.logging_conf

        container = ServiceContainer(
            service_cls=service_cls,
            version="0.0.0",
            config_paths=[],
            disable_service_registration=True,
            config=config,
        )

        await container.setup()
        await container.start()
        return container, container.service_instance

    # Inputs and Outputs management.
    def get_events(self, mock_name, wait_for_events=0, timeout=None):
        return self.get_service(mock_name).get_events(wait_for_events, timeout)

    def clean_events(self, mock_name):
        self.get_service(mock_name).clean_events()

    def add_http_response(self, mock_name, response):
        self.get_service(mock_name).add_http_response(response)

    def get_http_requests(self, mock_name):
        return self.get_service(mock_name).get_http_requests()

    def clean_http_requests(self, mock_name):
        self.get_service(mock_name).clean_http_requests()

    def add_remote_return_value(self, mock_name, value):
        self.get_service(mock_name).add_remote_return_value(value)

    def get_remote_calls(self, mock_name):
        return self.get_service(mock_name).get_remote_calls()

    def clean_remote_calls(self, mock_name):
        self.get_service(mock_name).clean_remote_calls()

    async def publish(self, event_type, obj, reply_to=None, extra_headers=None):
        if extra_headers is None:
            extra_headers = {}
        if not self._setup:
            await self.messenger.setup()
            self._setup = True
        logging.info("Publishing message %s - %s", event_type, obj)
        try:
            await self.messenger.publish(event_type, obj, reply_to, extra_headers)
            logging.info("Message published successfully.")
        except Exception as e:
            logging.error("Error while publishing message: %s", e)
            raise

    async def call(
        self, service_name, method_name, *args, extra_headers=None, **kwargs
    ):
        if not self._setup:
            await self.messenger.setup()
            self._setup = True
        logging.info("Calling RPC %s.%s", service_name, method_name)
        try:
            if extra_headers is None:
                extra_headers = {}
            return await self.messenger.call(
                service_name, method_name, *args, extra_headers=extra_headers, **kwargs
            )
        except Exception as e:
            logging.error("Error while calling RPC: %s", e)
            raise

    def get_service(self, mock_name):
        if mock_name not in self.mocks:
            raise KeyError(
                f"Unknown mock service: {mock_name!r}. Known: {list(self.mocks)}"
            )
        _, service = self.mocks[mock_name]
        logging.info("Getting service %s: %s", mock_name, service)
        return service



from aiohttp import web

from kontiki.messaging.consumer.rpc import rpc, rpc_error
from kontiki.registry.server.core import ServiceRegistryCore, ServiceStatus
from kontiki.web.web import http

# -----------------------------------------------------------------------------


class ServiceRegistry:
    core = ServiceRegistryCore()

    # -----------------------------------------------------------------------------
    # HTTP API
    # -----------------------------------------------------------------------------
    @http("/services", "GET")
    async def http_get_all_services(self, request):
        services = self.core.get_services()
        return web.json_response(services)

    @http("/services/{status}", "GET")
    async def http_get_services_by_status(self, request, status):
        try:
            _ = ServiceStatus(status)
        except ValueError as err:
            raise web.HTTPBadRequest(reason=f"Invalid status: {status}") from err

        services = self.core.get_services(status)
        return web.json_response(services)

    @http("/events", "GET")
    async def http_get_events(self, request):
        return web.json_response(self.core.get_events())

    @http("/events/{filter_field}/{value}", "GET")
    async def http_get_filtered_events(self, request, filter_field, value):
        matching_events = [
            event
            for event in self.core.get_events()
            if event.get(filter_field) == value
        ]
        return web.json_response(matching_events)

    @http("/exceptions", "GET")
    async def http_get_exceptions(self, request):
        return web.json_response(self.core.get_exceptions())

    @http("/exceptions/{filter_field}/{value}", "GET")
    async def http_get_filtered_exceptions(self, request, filter_field, value):
        matching_exceptions = [
            exception
            for exception in self.core.get_exceptions()
            if exception.get(filter_field) == value
        ]
        return web.json_response(matching_exceptions)

    # -----------------------------------------------------------------------------
    # AMQP API
    # -----------------------------------------------------------------------------

    @rpc
    async def get_services(self, status=None):
        if status is not None:
            try:
                _ = ServiceStatus(status)
            except ValueError:
                return rpc_error("INVALID_STATUS", f"Invalid status: {status}")
        return self.core.get_services(status)

    @rpc
    async def get_events(self):
        return self.core.get_events()

    @rpc
    async def get_filtered_events(self, filter_field, value):
        matching_events = [
            event
            for event in self.core.get_events()
            if event.get(filter_field) == value
        ]
        return matching_events

    @rpc
    async def get_exceptions(self):
        return self.core.get_exceptions()

    @rpc
    async def get_filtered_exceptions(self, filter_field, value):
        matching_exceptions = [
            exception
            for exception in self.core.get_exceptions()
            if exception.get(filter_field) == value
        ]
        return matching_exceptions



import logging
import threading
import time

from kontiki.delegate import ServiceDelegate

# -----------------------------------------------------------------------------


class EventManager(ServiceDelegate):
    def __init__(self):
        self.events = []
        self._lock = threading.Lock()
        self._condition = threading.Condition(self._lock)
        super().__init__()

    def store_event(self, event):
        logging.info("Storing event %s.", event)
        with self._condition:
            self.events.append(event)
            self._condition.notify_all()

    def get_events(self, wait_for_events, timeout=None):
        with self._condition:
            if wait_for_events <= 0:
                return self.events.copy()
            start_time = time.time()
            while len(self.events) < wait_for_events:
                if timeout and (time.time() - start_time) > timeout:
                    logging.warning(
                        "Timeout waiting for %s events after %.2f seconds, got %s.",
                        wait_for_events,
                        timeout,
                        len(self.events),
                    )
                    break
                logging.info(
                    "Waiting for %s events, currently have %s.",
                    wait_for_events,
                    len(self.events),
                )
                remaining = timeout - (time.time() - start_time) if timeout else 1.0
                wait_time = min(1.0, max(0.0, remaining))
                self._condition.wait(timeout=wait_time)

            if not self.events:
                logging.warning("No events stored.")
                return []
            return self.events.copy()

    def clean(self):
        logging.info("Cleaning events.")
        with self._lock:
            self.events.clear()



import uuid

from aio_pika import connect_robust

from kontiki.configuration.parameter import (
    get_kontiki_parameter,
    resolve_parameter_path,
)
from kontiki.messaging.common import (
    KONTIKI_SESSION_OPEN_RPC,
    create_tls_context,
    declare_event_exchange,
    declare_rpc_exchange,
    get_amqp_url,
)
from kontiki.messaging.consumer.event import OnEventTask
from kontiki.messaging.consumer.rpc import RpcTask
from kontiki.messaging.serialization import Serializer
from kontiki.utils import log


class Consumer:
    def __init__(self, container):
        self.container = container
        self.amqp_url = get_amqp_url(self.container.config)
        self.on_event_tasks = []
        self.rpc_tasks = []
        self.service_name = self.container.service_name
        self.connection = None

    async def setup(self):
        # Connect to RabbitMQ
        # Create SSL context if configured.
        tls_ctx = create_tls_context(self.container.config)
        self.connection = await connect_robust(self.amqp_url, ssl_context=tls_ctx)
        self.channel = await self.connection.channel()
        # Set QoS for AMQP channel
        prefetch_count = get_kontiki_parameter(
            self.container.config, "amqp.max_pending_messages", 10
        )
        await self.channel.set_qos(prefetch_count=prefetch_count)
        self.event_exchange = await declare_event_exchange(self.channel)
        self.rpc_exchange = await declare_rpc_exchange(self.channel)

        # Sets serializer
        self.serializer = Serializer(self.container.config)

        # Register internal Kontiki RPCs
        await self._add_internal_session_rpc()

    async def start(self):
        for task in self.on_event_tasks:
            await task.run()
        for task in self.rpc_tasks:
            await task.run()

    async def stop(self):
        if self.connection:
            await self.connection.close()

    async def _add_internal_session_rpc(self):
        """Register the internal Kontiki RPC used to open sessions."""

        async def _session_open_handler(container, _headers=None, **kwargs):
            session_id = str(uuid.uuid4())
            instance_id = str(container.instance_id)
            log.debug(
                "Session opened for service instance %s with ID: %s",
                instance_id,
                session_id,
            )
            return (instance_id, session_id)

        _session_open_handler._rpc_endpoint = {
            "name": KONTIKI_SESSION_OPEN_RPC,
            "include_headers": True,
        }

        await self.add_rpc_tasks([_session_open_handler])

    async def add_on_event_tasks(self, tasks):
        for task in tasks:
            task_data = task._on_event_endpoint
            event_type_or_key = task_data["event_type_or_key"]
            use_config = task_data["use_config"]
            target_instance = task_data["target_instance"]
            include_headers = task_data["include_headers"]
            requeue_on_error = task_data["requeue_on_error"]
            reject_on_redelivered = task_data["reject_on_redelivered"]
            broadcast = task_data.get("broadcast", False)

            # if use_config, search for the event_type in the conf from the path
            event_type = resolve_parameter_path(
                self.container.config, event_type_or_key, use_config
            )

            # Explicitly bind the task to the service instance
            # pylint: disable=unnecessary-dunder-call
            task = task.__get__(self.container.service_instance)

            # Declare queue
            if broadcast:
                # One queue per instance so that all instances receive the event.
                qname = (
                    f"{self.service_name}.{event_type}."
                    f"{self.container.instance_id}.queue"
                )
            else:
                # Single queue per service (competing consumers within the service).
                qname = f"{self.service_name}.{event_type}.queue"
            queue = await self.channel.declare_queue(qname, durable=True)

            # Define routing key regarding the target_instance param.
            # For per-instance (session) handlers we suffix with the
            # service instance_id.
            routing_key = (
                f"{event_type}.{self.container.instance_id}"
                if target_instance
                else event_type
            )
            # Bind queue to exchange with event type as the routing key
            await queue.bind(self.event_exchange, routing_key=routing_key)
            log.debug("Queue %s bound with routing key %s", qname, routing_key)

            # Declare and register the task
            on_event_task = OnEventTask(
                event_type,
                task,
                queue,
                self.serializer,
                include_headers,
                requeue_on_error,
                reject_on_redelivered,
            )
            self.on_event_tasks.append(on_event_task)
            log.debug("On event task registered for event: %s", event_type)

    async def add_rpc_tasks(self, tasks):
        for task in tasks:
            endpoint = task._rpc_endpoint
            task_name = endpoint["name"]
            include_headers = endpoint["include_headers"]

            # Declare queue
            routing_key = f"{self.service_name}.{task_name}"
            qname = f"{routing_key}.queue"
            queue = await self.channel.declare_queue(qname, durable=True)

            # Bind queue to exchange with event type as the routing key
            await queue.bind(self.rpc_exchange, routing_key=routing_key)
            log.debug("Queue %s bound with routing key %s", qname, routing_key)

            # Declare and register the task
            reply_exchange = self.channel.default_exchange
            rpc_task = RpcTask(
                task_name,
                task,
                self.container,
                reply_exchange,
                queue,
                self.serializer,
                include_headers,
            )
            self.rpc_tasks.append(rpc_task)
            log.debug("RPC task registered for task: %s", task_name)



from kontiki import __version__ as kontiki_version
from kontiki.registry.server.service import ServiceRegistry
from kontiki.runner import cli


def run():
    cli.run(
        ServiceRegistry,
        "Service registry for kontiki services.",
        version=kontiki_version,
        disable_service_registration=True,
    )



import logging
from datetime import datetime, timedelta

from aio_pika import Message

from kontiki.registry.common import HEARTBEAT_RKEY

# -----------------------------------------------------------------------------


class HeartbeatManager:
    def __init__(self, core):
        self.core = core
        self.heartbeats = {}
        self.degraded_services = []
        self.started_at = datetime.now()

    async def setup(self):
        # Declare queues
        await self.core.create_and_consume_queue(HEARTBEAT_RKEY, self._handle_heartbeat)

    async def _handle_heartbeat(self, message):
        async with message.process():
            logging.debug("Received heartbeat message: %s", message.body)
            try:
                data = self.core.serializer.loads(message.body)
                service_name = data["service_name"]
                instance_id = data["instance_id"]
                degraded = data["degraded"]
                await self.store_heartbeat(service_name, instance_id, degraded)

            except Exception as e:
                logging.error("Error processing heartbeat: %s", e)

    async def store_heartbeat(self, service_name, instance_id, degraded):
        logging.debug("Stored heartbeats = %s", self.heartbeats)
        try:
            service_str = f"{service_name}#{instance_id}"

            service_instance = (service_name, instance_id)
            if self.core.registry.has_service_instance(service_name, instance_id):
                if degraded and service_instance not in self.degraded_services:
                    self.degraded_services.append(service_instance)
                    logging.warning("%s enters degraded state.", service_str)
                else:
                    if not degraded:
                        if service_instance in self.degraded_services:
                            logging.info(
                                "%s recovers from degraded state.", service_str
                            )
                            self.degraded_services.remove(service_instance)
                self.heartbeats[service_instance] = datetime.now()
                logging.debug("Heartbeat updated for %s.", service_str)
            else:
                # We dont ask for registering again when we just start. We might have
                # received a heartbeat before the service registry is ready.
                grace_period = timedelta(seconds=15)
                if datetime.now() - self.started_at < grace_period:
                    logging.warning(
                        "Heartbeat for unknown service %s received during"
                        " grace period, ignoring for now.",
                        service_str,
                    )
                    return
                logging.warning(
                    "Heartbeat received for unknown service %s", service_str
                )
                registry_admin_exchange = self.core.registry_admin_exchange
                rkey = f"{service_name}.{instance_id}.register_again"
                message = Message(body=b"")
                await registry_admin_exchange.publish(message, routing_key=rkey)
                logging.info("Published register_again message to %s", rkey)

        except Exception as e:
            logging.error("Heartbeat Storage Error : %s", e)

    def get_last_heartbeat(self, service_name, instance_id):
        return self.heartbeats.get((service_name, instance_id), None)

    def is_degraded(self, service_name, instance_id):
        return (service_name, instance_id) in self.degraded_services



from kontiki.utils import log


def extract_schema(model_class):
    try:
        # Pydantic v2
        if hasattr(model_class, "model_json_schema"):
            return model_class.model_json_schema()  # type: ignore[attr-defined]
        # Pydantic v1
        if hasattr(model_class, "schema"):
            return model_class.schema()  # type: ignore[attr-defined]
        # Fallback
        return {"type": "object", "description": "Custom model"}
    except Exception as exc:  # pragma: no cover - defensive
        log.warning("Could not extract schema from %s: %s", model_class, exc)
        return {"type": "object", "description": "Model schema extraction failed"}


def parse_with_model(model_class, data):
    """Parse/validate data using a Pydantic model (v1 or v2)."""
    if hasattr(model_class, "model_validate"):
        return model_class.model_validate(data)  # type: ignore[attr-defined]
    if hasattr(model_class, "parse_obj"):
        return model_class.parse_obj(data)  # type: ignore[attr-defined]
    # If unknown type, just return data
    return data



import logging

from kontiki.registry.common import REGISTER_RKEY, UNREGISTER_RKEY, nested_dict

# -----------------------------------------------------------------------------


class Registry:
    def __init__(self, core):
        self.core = core
        self.services = nested_dict()

    async def setup(self):
        # Declare queues
        handlers = {
            REGISTER_RKEY: self._handle_register,
            UNREGISTER_RKEY: self._handle_unregister,
        }

        for rkey, callback in handlers.items():
            await self.core.create_and_consume_queue(rkey, callback)

    async def _handle_register(self, message):
        async with message.process():
            logging.info("Received registration message: %s", message.body)
            try:
                data = self.core.serializer.loads(message.body)
                service_name = data["service_name"]
                instance_id = data["instance_id"]

                # Services registration
                self.services[service_name][instance_id] = data
                logging.info("Registering service %s#%s", service_name, instance_id)

            except Exception as e:
                logging.error("Error processing registration: %s", e)

    async def _handle_unregister(self, message):
        async with message.process():
            logging.info("Received unregistration message: %s", message.body)
            try:
                data = self.core.serializer.loads(message.body)
                service_name = data["service_name"]
                instance_id = data["instance_id"]

                if service_name not in self.services:
                    logging.error("Service %s not found.", service_name)
                    return

                if instance_id not in self.services[service_name]:
                    logging.error(
                        "Instance %s of service %s not found.",
                        instance_id,
                        service_name,
                    )
                    return

                del self.services[service_name][instance_id]

                # No more service instances, service entry removal.
                if not self.services[service_name]:
                    logging.info(
                        "No active instances for %s. Cleaning up...", service_name
                    )
                    # Service entry removal.
                    del self.services[service_name]

            except Exception as e:
                logging.error("Error processing unregistration: %s", e)

    def has_service(self, service_name):
        return service_name in self.services

    def has_service_instance(self, service_name, instance_id):
        has_service = self.has_service(service_name)
        return has_service and instance_id in self.services[service_name]



# flake8: noqa


def build_swagger_ui_html(title, spec_url, description=""):
    return f"""
<!DOCTYPE html>
<html>
<head>
  <meta charset=\"utf-8\">
  <title>{title} - Swagger UI</title>
  <meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">
  <link rel=\"stylesheet\" href=\"https://unpkg.com/swagger-ui-dist@5/swagger-ui.css\">
  <style> body {{ margin:0; background:#fafafa; }} .topbar {{ display:none; }} </style>
  <meta name=\"description\" content=\"{description}\">
  <script>window.specUrl = '{spec_url}';</script>
  <script src=\"https://unpkg.com/swagger-ui-dist@5/swagger-ui-bundle.js\"></script>
  <script src=\"https://unpkg.com/swagger-ui-dist@5/swagger-ui-standalone-preset.js\"></script>
  <script>
  window.onload = function() {{
    window.ui = SwaggerUIBundle({{
      url: window.specUrl,
      dom_id: '#swagger-ui',
      presets: [SwaggerUIBundle.presets.apis, SwaggerUIStandalonePreset],
      layout: 'BaseLayout'
    }});
  }}
  </script>
  </head>
  <body>
    <div id=\"swagger-ui\"></div>
  </body>
</html>
"""



from kontiki.utils import log


class ServiceDelegate:
    def __init__(self):
        self.container = None

    def bind(self, container, attr_name):
        self.container = container
        self.attr_name = attr_name

    async def setup(self):
        pass

    async def start(self):
        pass

    async def stop(self):
        pass

    async def publish_exception(self, exception, context=None):
        if context is None:
            context = {}

        if not self.container or not self.container.service_registry_client:
            log.warning("Service registration is disabled or unavailable..")
            return

        try:
            await self.container.service_registry_client.register_exception(
                exception, context
            )
            log.debug("Exception published successfully: %s.", exception)
        except Exception as e:
            log.error("Error while publishing exception %s: %s", exception, e)



import yaml

from kontiki.utils import log

# -----------------------------------------------------------------------------


class ConfigMergeError(Exception):
    pass


def merge(config_files):
    def _smart_merge(d1, d2):
        shared_keys = d1.keys() & d2.keys()
        only_keys_2 = d2.keys() - d1.keys()
        for key in only_keys_2:
            d1[key] = d2[key]
        errors = []
        warnings = []

        for key in shared_keys:
            if isinstance(d1[key], dict) and isinstance(d2[key], dict):
                suberrors, subwarnings = _smart_merge(d1[key], d2[key])
                subwarnings = [f"{key}:{subkey}" for subkey in subwarnings]
                suberrors = [f"{key}:{subkey}" for subkey in suberrors]

                errors += suberrors
                warnings += subwarnings

            else:
                if d1[key] == d2[key]:
                    warnings.append(key)
                else:
                    errors.append(key)

        return errors, warnings

    config = {}
    for config_file in config_files:
        with open(config_file, "r", encoding="utf-8") as file:
            tmp_dict = yaml.load(file, Loader=yaml.FullLoader)

            if tmp_dict is None:
                tmp_dict = {}
            errors, warnings = _smart_merge(config, tmp_dict)
            if errors:
                for param in errors:
                    log.error("%s definition is inconsistent.", param)
                msg = (
                    "Configuration file merge has failed due to param"
                    " definition inconsistency."
                )
                raise ConfigMergeError(msg)
            if warnings:
                for param in warnings:
                    msg = "%s is defined twice with the same value."
                    log.warning(msg, param)

    return config



import inspect

from aiohttp import web

from kontiki.configuration.parameter import (
    get_kontiki_parameter,
    resolve_parameter_path,
)
from kontiki.utils import log
from kontiki.web.documentation import (
    get_http_documentation,
    register_auto_docs_endpoints,
)
from kontiki.web.utils import extract_schema, parse_with_model

# -----------------------------------------------------------------------------


def http(
    path_or_key,
    method,
    use_config=False,
    version=None,
    summary=None,
    description=None,
    tags=None,
    request_model=None,
    response_model=None,
    status_code=200,
    responses=None,
    errors=None,
    skip_documentation=False,
    validate_request=False,
):
    def decorator(handler):
        handler._http_endpoint = (path_or_key, method, use_config)
        request_schema = None
        response_schema = None

        if request_model and inspect.isclass(request_model):
            request_schema = extract_schema(request_model)
        elif isinstance(request_model, dict):
            request_schema = request_model

        if response_model and inspect.isclass(response_model):
            response_schema = extract_schema(response_model)
        elif isinstance(response_model, dict):
            response_schema = response_model

        handler._http_documentation = {
            "version": version,
            "summary": summary,
            "description": description,
            "tags": tags or [],
            "request_model": request_model,
            "request_schema": request_schema,
            "response_model": response_model,
            "response_schema": response_schema,
            "status_code": status_code,
            "responses": responses or {},
            "errors": tuple(errors) if errors else (),
            "skip_documentation": skip_documentation,
            "validate_request": validate_request,
        }

        return handler

    return decorator


# -----------------------------------------------------------------------------


class HttpServer:
    def __init__(self, container, entrypoints):
        self.app = web.Application()
        self.entrypoints = entrypoints
        self.container = container
        self._openapi_cache = {}
        self._docs_registered = False
        self.add_entrypoints(entrypoints)

    async def setup(self):
        if not self._docs_registered:
            try:
                register_auto_docs_endpoints(
                    app=self.app,
                    container=self.container,
                    entrypoints=self.entrypoints,
                    openapi_cache=self._openapi_cache,
                )
                self._docs_registered = True
            except Exception as e:
                log.error("Error registering auto documentation endpoints: %s", e)

        self.runner = web.AppRunner(self.app)
        await self.runner.setup()

    async def start(self):
        address = get_kontiki_parameter(
            self.container.config, "http.address", "0.0.0.0"
        )
        port = get_kontiki_parameter(self.container.config, "http.port", 8080)
        self.site = web.TCPSite(self.runner, address, port)
        log.info("Service running on http://%s:%s", address, port)
        await self.site.start()

    async def stop(self):
        log.info("Stopping HTTP server...")
        try:
            if self.site:
                await self.site.stop()
            if self.runner:
                await self.runner.cleanup()
            log.info("HTTP server stopped.")
        except Exception as e:
            log.error("Error while stopping HTTP server: %s", e)

    def add_entrypoints(self, entrypoints):
        for handler in entrypoints:
            path_or_key, method, use_config = handler._http_endpoint
            path = resolve_parameter_path(
                self.container.config, path_or_key, use_config
            )
            log.debug(
                "Registering route: %s %s with handler %s",
                method,
                path,
                handler.__name__,
            )

            # Bind the handler to the service instance
            # pylint: disable=unnecessary-dunder-call
            bound_handler = handler.__get__(self.container.service_instance)

            # Create the handler
            route_handler = self._create_handler(bound_handler, path, method, handler)

            # Register the route
            try:
                self.app.router.add_route(method, path, route_handler)
                msg = "Route %s %s successfully registered with %s"
                log.info(msg, method, path, handler.__name__)
            except Exception as e:
                log.error("Error registering route %s %s: %s", method, path, e)

    def _create_handler(self, bound_handler, path, method, original_handler):
        service_class = type(self.container.service_instance)
        error_handlers = getattr(service_class, "http_error_handlers", None) or {}

        async def handler(request, *args, **kwargs):
            msg = "Handling %s %s with params: %s"
            log.debug(msg, method, path, request.match_info)
            doc = getattr(original_handler, "_http_documentation", {}) or {}
            validate_request = bool(doc.get("validate_request"))
            request_model = doc.get("request_model")
            parsed_body = None

            try:
                if (
                    validate_request
                    and request_model
                    and method.upper() in ("POST", "PUT", "PATCH")
                ):
                    if inspect.isclass(request_model):
                        try:
                            data = await request.json()
                            parsed_body = parse_with_model(request_model, data)
                        except Exception as e:
                            log.warning("Invalid request body: %s", e)
                            raise web.HTTPUnprocessableEntity(
                                reason="Invalid request body"
                            )
                    elif isinstance(request_model, dict):
                        # No validation required for dict-based models
                        parsed_body = await request.json()

                call_kwargs = dict(request.match_info)
                if parsed_body is not None:
                    call_kwargs["body"] = parsed_body

                response = await bound_handler(request, **call_kwargs)
                if isinstance(response, web.StreamResponse):
                    return response

                if isinstance(response, (dict, list)):
                    return web.json_response(
                        response, status=doc.get("status_code", 200)
                    )

                return response
            except Exception as e:
                mapped = self._resolve_http_error_mapping(type(e), error_handlers)
                if mapped is not None:
                    status, message_override = mapped
                    body = (
                        message_override
                        if message_override is not None
                        else getattr(e, "message", None) or str(e)
                    )
                    log.warning(
                        "Mapped exception in handler for %s %s: %s", method, path, e
                    )
                    return web.json_response({"message": body}, status=status)

                log.error("Error in handler for %s %s: %s", method, path, e)
                raise web.HTTPInternalServerError(reason="Internal Server Error") from e

        return handler

    @staticmethod
    def _resolve_http_error_mapping(exception_type, handlers_map):
        if not handlers_map:
            return None
        for cls in inspect.getmro(exception_type):
            if cls in handlers_map:
                val = handlers_map[cls]
                if isinstance(val, (tuple, list)) and len(val) >= 2:
                    return (int(val[0]), val[1])
                if isinstance(val, (tuple, list)) and len(val) == 1:
                    return (int(val[0]), None)
                return None
        return None

    def get_documentation(self, version=None):
        return get_http_documentation(self.container, self.entrypoints, version=version)



#!/usr/bin/env python3
"""
Main entry point for kontiki run command.
Usage: kontiki <service_class> --config <config_file.yaml>
"""

import argparse
import asyncio
import importlib
import sys
from pathlib import Path

from kontiki.runner.runner import run_service


def import_service_class(service_class_path):
    try:
        module_path, class_name = service_class_path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        service_class = getattr(module, class_name)
        return service_class
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Cannot import service class '{service_class_path}': {e}")


def main():
    parser = argparse.ArgumentParser(
        description="Run a Kontiki service",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "service_class",
        help="Full path to the service class (e.g., 'myapp.services.MyService')",
    )

    parser.add_argument(
        "--config",
        action="append",
        required=True,
        help="Path to a config file. Specify multiple times for multiple files.",
    )

    parser.add_argument(
        "--version", default="1.0.0", help="Service version (default: 1.0.0)"
    )

    parser.add_argument(
        "--disable-service-registration",
        action="store_true",
        help="Disable automatic service registration",
    )

    args = parser.parse_args()

    # Validate config files exist
    for config_path in args.config:
        if not Path(config_path).exists():
            print(f"Error: Config file '{config_path}' does not exist", file=sys.stderr)
            sys.exit(1)

    try:
        # Import the service class
        service_class = import_service_class(args.service_class)

        # Run the service
        asyncio.run(
            run_service(
                service_class,
                args.config,
                args.version,
                args.disable_service_registration,
            )
        )
    except ImportError as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        print("\nService stopped by user", file=sys.stderr)
        sys.exit(0)
    except Exception as e:
        print(f"Error running service: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



from aiohttp import web

from kontiki.configuration.parameter import (
    get_kontiki_parameter,
    resolve_parameter_path,
)
from kontiki.utils import log


def get_http_documentation(container, entrypoints, version=None):
    service_class = type(container.service_instance)
    error_handlers = getattr(service_class, "http_error_handlers", None) or {}

    docs = []
    for handler in entrypoints:
        if not hasattr(handler, "_http_documentation"):
            continue

        path_or_key, method, use_config = handler._http_endpoint
        path = resolve_parameter_path(container.config, path_or_key, use_config)

        doc = {
            "path": path,
            "method": method,
            "handler_name": handler.__name__,
            **getattr(handler, "_http_documentation", {}),
        }

        errors = doc.get("errors") or ()
        derived_responses = {}
        for exc_type in errors:
            if exc_type in error_handlers:
                val = error_handlers[exc_type]
                if isinstance(val, (tuple, list)) and len(val) >= 1:
                    status = int(val[0])
                    desc = (
                        val[1]
                        if len(val) >= 2 and val[1] is not None
                        else exc_type.__name__
                    )
                    derived_responses[status] = desc
        doc["responses"] = {**derived_responses, **doc.get("responses", {})}

        if doc.get("skip_documentation"):
            continue

        if version is not None and doc.get("version") != version:
            continue

        docs.append(doc)

    return docs


def _detect_api_versions(entrypoints):
    versions = set()
    for handler in entrypoints:
        if hasattr(handler, "_http_documentation"):
            doc = getattr(handler, "_http_documentation", {}) or {}
            v = doc.get("version")
            if v:
                versions.add(v)
    return versions


def _build_openapi_from_endpoints(endpoints, title, version, description):
    openapi_doc = {
        "openapi": "3.0.0",
        "info": {"title": title, "description": description, "version": version},
        "paths": {},
        "tags": [],
    }

    all_tags = set()
    all_defs = {}

    def _collect_defs_from_schema(schema):
        if isinstance(schema, dict):
            cleaned_schema = dict(schema)
            if "$defs" in cleaned_schema:
                defs = cleaned_schema.get("$defs") or {}
                if isinstance(defs, dict):
                    all_defs.update(defs)
                cleaned_schema.pop("$defs", None)
            for key, value in list(cleaned_schema.items()):
                if isinstance(value, (dict, list)):
                    cleaned_schema[key] = _collect_defs_from_schema(value)
                elif (
                    key == "$ref"
                    and isinstance(value, str)
                    and value.startswith("#/$defs/")
                ):
                    def_name = value.replace("#/$defs/", "")
                    cleaned_schema[key] = f"#/components/schemas/{def_name}"
            return cleaned_schema
        if isinstance(schema, list):
            return [_collect_defs_from_schema(item) for item in schema]
        return schema

    for endpoint in endpoints:
        if not isinstance(endpoint, dict):
            continue

        path = endpoint.get("path")
        method = str(endpoint.get("method", "get")).lower()
        if not path or not method:
            continue

        if path not in openapi_doc["paths"]:
            openapi_doc["paths"][path] = {}

        operation = {
            "summary": endpoint.get("summary") or f"{method.upper()} {path}",
            "description": endpoint.get("description", ""),
            "tags": endpoint.get("tags", []) or [],
            "responses": {},
        }

        responses = endpoint.get("responses") or {}
        success_status = str(endpoint.get("status_code", 200))
        if success_status not in responses:
            operation["responses"][success_status] = {
                "description": "Successful response"
            }

        request_schema = endpoint.get("request_schema")
        if request_schema and method in ["post", "put", "patch"]:
            cleaned_request_schema = _collect_defs_from_schema(request_schema)
            operation["requestBody"] = {
                "content": {"application/json": {"schema": cleaned_request_schema}},
                "required": True,
            }

        response_schema = endpoint.get("response_schema")
        if response_schema:
            cleaned_response_schema = _collect_defs_from_schema(response_schema)
            status_code = str(endpoint.get("status_code", 200))
            operation["responses"].setdefault(
                status_code, {"description": "Successful response"}
            )
            operation["responses"][status_code]["content"] = {
                "application/json": {"schema": cleaned_response_schema}
            }

        for sc, resp in responses.items():
            sc_str = str(sc)
            if isinstance(resp, str):
                operation["responses"][sc_str] = {"description": resp}
            elif isinstance(resp, dict):
                operation["responses"][sc_str] = resp

        openapi_doc["paths"][path][method] = operation
        all_tags.update(operation["tags"])

    openapi_doc["tags"] = [{"name": tag} for tag in sorted(all_tags)]
    if all_defs:
        openapi_doc["components"] = {"schemas": all_defs}

    return openapi_doc


def register_auto_docs_endpoints(app, container, entrypoints, openapi_cache):
    enable = get_kontiki_parameter(container.config, "http.documentation.enabled", True)
    if not enable:
        log.info("Auto docs endpoints disabled by configuration.")
        return

    path_template = get_kontiki_parameter(
        container.config,
        "http.documentation.path_template",
        "/api/{version}/docs",
    )

    versions = _detect_api_versions(entrypoints)
    for version in versions:
        _register_docs_endpoint(
            app=app,
            container=container,
            entrypoints=entrypoints,
            openapi_cache=openapi_cache,
            version=version,
            path_template=path_template,
        )


def _register_docs_endpoint(
    app, container, entrypoints, openapi_cache, version, path_template
):
    path = path_template.format(version=version)

    if path.endswith("/docs"):
        openapi_path = path[:-5] + "/openapi.json"
    else:
        openapi_path = f"{path}.json"

    doc_title = get_kontiki_parameter(
        container.config,
        "http.documentation.title",
        container.service_name,
    )
    doc_desc = get_kontiki_parameter(
        container.config,
        "http.documentation.description",
        f"API documentation for {container.service_name}",
    )

    async def openapi_handler(request):
        cached = openapi_cache.get(version)
        if cached is not None:
            return web.json_response(cached)

        endpoints = get_http_documentation(container, entrypoints, version=version)
        openapi = _build_openapi_from_endpoints(
            endpoints=endpoints,
            title=str(doc_title),
            version=str(version),
            description=str(doc_desc),
        )
        openapi_cache[version] = openapi
        return web.json_response(openapi)

    async def docs_handler(request):
        default_ui = "swagger"
        target = f"{path}/{default_ui}"
        raise web.HTTPFound(location=target)

    try:
        app.router.add_route("GET", path, docs_handler)
        app.router.add_route("GET", openapi_path, openapi_handler)
        log.info("Auto-registered docs endpoint: GET %s (redirect)", path)
        log.info("Auto-registered OpenAPI endpoint: GET %s", openapi_path)
    except Exception as exc:
        log.error(
            "Error registering docs/openapi endpoints %s / %s: %s",
            path,
            openapi_path,
            exc,
        )

    swagger_path = f"{path}/swagger"

    async def swagger_handler(request):
        spec_url = openapi_path
        from .swagger import build_swagger_ui_html

        html = build_swagger_ui_html(
            title=f"{doc_title} ({version})",
            spec_url=spec_url,
            description=doc_desc,
        )
        return web.Response(text=html, content_type="text/html")

    try:
        app.router.add_route("GET", swagger_path, swagger_handler)
        log.info("Auto-registered Swagger UI endpoint: GET %s", swagger_path)
    except Exception as exc:
        log.error("Error registering Swagger UI endpoint %s: %s", swagger_path, exc)



import logging

from kontiki.delegate import ServiceDelegate


class HttpManager(ServiceDelegate):
    def __init__(self):
        self.requests = []
        self.responses = []
        super().__init__()

    def add_response(self, response):
        logging.info("Adding http response %s (%s)", response, self)
        self.responses.append(response)

    def store_request(self, payload):
        logging.info("Storing request %s (%s)", payload, self)
        self.requests.append(payload)

    def get_requests(self):
        if not self.requests:
            logging.warning("No HTTP requests stored.")
        return self.requests

    def get_response(self):
        if self.responses:
            return self.responses.pop(0)
        raise RuntimeError("No more responses to return.")

    def clean(self):
        self.requests.clear()
        self.responses.clear()



from kontiki.utils import get_kontiki_header_name


class EventSession:
    def __init__(self, messenger, service_name, instance_id, session_id):
        self._messenger = messenger
        self.service_name = service_name
        self.instance_id = str(instance_id)
        self.session_id = str(session_id)

    async def publish(self, event_type, obj, extra_headers=None):
        if extra_headers is None:
            extra_headers = {}

        routing_key = f"{event_type}.{self.instance_id}"
        headers = {
            get_kontiki_header_name("session_id"): self.session_id,
            **extra_headers,
        }
        await self._messenger.publish(routing_key, obj, extra_headers=headers)



import asyncio
import logging
from datetime import datetime, timedelta, timezone

from kontiki.configuration.parameter import get_parameter
from kontiki.messaging.common import declare_event_exchange, declare_rpc_exchange
from kontiki.registry.common import declare_registry_event_exchange
from kontiki.utils import get_kontiki_prefix

# -----------------------------------------------------------------------------


class EventTracker:
    def __init__(self, core):
        self.core = core
        self.consumers = {}
        self.events = []

    async def setup(self):
        self.event_ttl_minutes = get_parameter(
            self.core.container.config, "event_tracker.ttl_minutes", 0
        )

        self.event_ttl_hours = get_parameter(
            self.core.container.config, "event_tracker.ttl_hours", 24 * 7
        )

        # Use minutes if set, otherwise use hours.
        if self.event_ttl_minutes > 0:
            self.event_ttl = self.event_ttl_minutes
        elif self.event_ttl_hours > 0:
            self.event_ttl = self.event_ttl_hours * 60

        self.disable_tracking = get_parameter(
            self.core.container.config, "event_tracker.disable", False
        )

        self.cleanup_interval = get_parameter(
            self.core.container.config, "event_tracker.cleanup_interval_seconds", 3600
        )

        if self.is_disabled():
            return

        queue_name = "event_tracker.queue"
        try:
            self.registry_event_exchange = await declare_registry_event_exchange(
                self.core.channel
            )
            event_exchange = await declare_event_exchange(self.core.channel)
            rpc_exchange = await declare_rpc_exchange(self.core.channel)
            await self.registry_event_exchange.bind(event_exchange, routing_key="#")
            await self.registry_event_exchange.bind(rpc_exchange, routing_key="#")
            queue = await self.core.channel.declare_queue(queue_name, durable=True)
            await queue.bind(self.registry_event_exchange, routing_key="#")
            await queue.consume(self._handle_event)

            self.cleanup_task = asyncio.create_task(self._cleanup_events())
            logging.debug("EventTracker setup completed.")
        except Exception as e:
            logging.error("Error creating or consuming queue %s: %s", queue_name, e)

    async def _handle_event(self, message):
        async with message.process():
            try:
                raw_headers = message.headers or {}

                # Normalise Kontiki-specific headers while preserving all headers.
                normalized = dict(raw_headers)
                prefix = get_kontiki_prefix()
                for key, value in raw_headers.items():
                    if key.startswith(prefix):
                        bare_key = key[len(prefix) :]
                        # Do not overwrite potential user headers with the same name.
                        normalized.setdefault(bare_key, value)

                event_type = normalized.get("event_type", "_rpc_event")
                service = normalized.get("service_name")
                uuid = normalized.get("instance_id")
                host = normalized.get("host")
                logging.debug(
                    "Received %s from %s#%s [%s]", event_type, service, uuid, host
                )
                self.events.append(normalized)

            except Exception as e:
                logging.error("Error processing event: %s", e)

    async def _cleanup_events(self):
        logging.info(
            "Starting cleanup task with %s seconds." "interval (TTL: %s minutes)",
            self.cleanup_interval,
            self.event_ttl,
        )

        while True:
            try:
                self._purge_expired_events()

                await asyncio.sleep(self.cleanup_interval)
            except asyncio.CancelledError:
                logging.info("Cleanup task cancelled.")
                break
            except Exception as e:
                logging.error("Error during cleanup task: %s", e)
                # Avoid a tight error loop that can spam logs and fill disk.
                await asyncio.sleep(self.cleanup_interval)

    def is_disabled(self):
        if self.disable_tracking:
            logging.info("Event tracking is disabled.")
        return self.disable_tracking

    def _purge_expired_events(self):
        expiration_time = datetime.now(timezone.utc) - timedelta(minutes=self.event_ttl)
        cutoff = 0
        for index, event in enumerate(self.events):
            event_timestamp = event.get("timestamp")
            if event_timestamp is None:
                cutoff = index + 1
                continue

            if isinstance(event_timestamp, str):
                try:
                    event_timestamp = datetime.fromisoformat(event_timestamp)
                except Exception:
                    logging.warning("Invalid timestamp format: %s", event_timestamp)
                    cutoff = index + 1
                    continue

            # Normalise to UTC-aware datetime to avoid naive/aware comparisons.
            if isinstance(event_timestamp, datetime):
                if event_timestamp.tzinfo is None:
                    event_timestamp = event_timestamp.replace(tzinfo=timezone.utc)
                else:
                    event_timestamp = event_timestamp.astimezone(timezone.utc)

            if event_timestamp <= expiration_time:
                cutoff = index + 1
                continue

            # Keep events that are not expired.
            break

        # Remove expired events.
        if cutoff > 0:
            del self.events[:cutoff]
            logging.debug("Cleaned up %s expired events.", cutoff)
        else:
            logging.debug("No expired events found.")



import logging

from kontiki.delegate import ServiceDelegate


class RpcManager(ServiceDelegate):
    def __init__(self):
        self.call_args = []
        self.return_values = []
        super().__init__()

    def add_return_value(self, value):
        logging.info("Adding return value %s", value)
        self.return_values.append(value)

    def store_call_args(self, *args, **kwargs):
        logging.info("Storing args %s and kwargs %s", args, kwargs)
        self.call_args.append((args, kwargs))

    def get_call_args(self):
        if not self.call_args:
            logging.warning("No args stored.")
        return self.call_args

    def get_return_value(self):
        logging.info("Getting return values.")
        try:
            if self.return_values:
                value = self.return_values.pop(0)
                logging.info("Got value %s.", value)
                return value
            raise RuntimeError("No more values to return.")
        except Exception as e:
            logging.error("Error while getting return_value %s", str(e))
            raise e

    def clean(self):
        self.call_args.clear()
        self.return_values.clear()



import ssl

from aio_pika import ExchangeType

from kontiki.configuration.parameter import get_kontiki_parameter
from kontiki.utils import KONTIKI

# -----------------------------------------------------------------------------

AMQP_DEFAULT_URL = "amqp://guest:guest@localhost/"
EVENT_EXCHANGE = "event_exchange"
RPC_EXCHANGE = "rpc_exchange"
KONTIKI_SESSION_OPEN_RPC = f"___{KONTIKI}__internal_session_open__"


async def declare_event_exchange(channel, name=EVENT_EXCHANGE):
    return await channel.declare_exchange(name, ExchangeType.TOPIC)


async def declare_rpc_exchange(channel, name=RPC_EXCHANGE):
    return await channel.declare_exchange(name, ExchangeType.TOPIC)


def get_amqp_url(config, default_url=AMQP_DEFAULT_URL):
    return get_kontiki_parameter(config, "amqp.url", default_url)


def get_rpc_timeout(config):
    return get_kontiki_parameter(config, "amqp.rpc.timeout", 10)


def create_tls_context(config):
    config = get_kontiki_parameter(config, "amqp.tls", {})

    if not isinstance(config, dict):
        return None

    if not config.get("enabled", False):
        return None

    context = ssl.create_default_context(cafile=config["ca_cert"])
    client_cert = config.get("client_cert", None)
    client_key = config.get("client_key", None)
    if client_cert and client_key:
        context.load_cert_chain(certfile=client_cert, keyfile=client_key)
    return context



from collections import defaultdict

from aio_pika import ExchangeType

from kontiki.configuration.parameter import get_kontiki_parameter

REGISTRY_ADMIN_EXCHANGE = "registry_admin_exchange"
REGISTRY_EVENT_EXCHANGE = "registry_event_exchange"
REGISTER_RKEY = "registry.register"
UNREGISTER_RKEY = "registry.unregister"
HEARTBEAT_RKEY = "registry.heartbeat"
EXCEPTION_RKEY = "registry.exception"
HEARTBEAT_DEFAULT_INTERVAL = 60


def get_event_queue_name(service_name, event):
    return f"{service_name}.{event}.queue"


async def declare_registry_admin_exchange(channel):
    return await channel.declare_exchange(REGISTRY_ADMIN_EXCHANGE, ExchangeType.TOPIC)


async def declare_registry_event_exchange(channel):
    return await channel.declare_exchange(REGISTRY_EVENT_EXCHANGE, ExchangeType.TOPIC)


def get_heartbeat_interval(config):
    return get_kontiki_parameter(
        config, "heartbeat.interval", HEARTBEAT_DEFAULT_INTERVAL
    )


def nested_dict():
    return defaultdict(nested_dict)



import json
import pickle

from kontiki.configuration.parameter import get_kontiki_parameter
from kontiki.messaging.rpc import RpcErrorType, RpcReturn
from kontiki.utils import log

# -----------------------------------------------------------------------------

DEFAULT_SERIALIZATION = "pickle"
SUPPORTED_SERIALIZATIONS = ["pickle", "json"]

# -----------------------------------------------------------------------------


class Serializer:
    def __init__(self, config, serialization=DEFAULT_SERIALIZATION):
        self.serialization = get_kontiki_parameter(
            config, "amqp.serialization", serialization
        )
        if self.serialization not in SUPPORTED_SERIALIZATIONS:
            msg = f"{self.serialization} serializer not supported."
            log.error(msg)
            raise RuntimeError(msg)

    def _rpcreturn_object_hook(self, obj):
        if obj.get("__rpcreturn__") is True:
            error_type_str = obj.get("error_type", "NONE")
            error_type = (
                RpcErrorType[error_type_str]
                if error_type_str in RpcErrorType.__members__
                else RpcErrorType.NONE
            )
            return RpcReturn(
                success=obj.get("success", False),
                result=obj.get("result"),
                message=obj.get("message"),
                error_type=error_type,
            )
        return obj

    def _default_encoder(self, obj):
        if isinstance(obj, RpcReturn):
            log.debug(
                "Encode RpcReturn object: %s",
                obj.result if obj.success else obj.message,
            )
            return {
                "__rpcreturn__": True,
                "success": obj.success,
                "result": obj.result,
                "message": obj.message,
                "error_type": obj.error_type.name,
            }
        raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")

    def loads(self, message_body):
        if self.serialization == "json":
            return json.loads(
                message_body.decode(), object_hook=self._rpcreturn_object_hook
            )

        return pickle.loads(message_body)

    def dumps(self, message_body):
        if self.serialization == "json":
            return json.dumps(message_body, default=self._default_encoder).encode()

        return pickle.dumps(message_body)



from .doc import get_http_documentation, register_auto_docs_endpoints
from .swagger import build_swagger_ui_html

__all__ = [
    "get_http_documentation",
    "register_auto_docs_endpoints",
    "build_swagger_ui_html",
]



from kontiki.utils import KONTIKI, log

# -----------------------------------------------------------------------------

_MISSING = object()


class ConfigParameterError(Exception):
    pass


class ConfigParameter:
    def __init__(self, path_or_key):
        self.path_or_key = path_or_key

    def get(self, config, default=_MISSING):
        keys = self.path_or_key.split(".")
        try:
            value = config
            for key in keys:
                value = value[key]
            return value
        except KeyError as err:
            if default is not _MISSING:
                msg = "No parameter %s defined in configuration. Use default value %s"
                log.info(msg, self.path_or_key, default)
                return default
            raise ConfigParameterError(
                f"Configuration paramater path {self.path_or_key} not found."
            ) from err


def resolve_parameter_path(config, path_or_key, use_config):
    if use_config:
        return ConfigParameter(path_or_key).get(config)
    return path_or_key


def get_kontiki_parameter(config, path, default=_MISSING):
    return get_parameter(config, f"{KONTIKI}.{path}", default)


def get_parameter(config, path, default=_MISSING):
    try:
        if config:
            return ConfigParameter(path).get(config, default)
        if default is not None:
            msg = "No parameter %s defined in configuration. Use default value %s"
            log.info(msg, path, default)

            return default
        msg = "No configuration found. No default value provided."
        log.error(msg)
        raise ConfigParameterError(msg)
    except Exception as e:
        log.error(str(e))
        raise e



class RpcError(Exception):
    def __init__(self, method_name, code, message):
        self.method_name = method_name
        self.code = code
        self.message = message
        super().__init__(f"{self.method_name} [{self.code}]: {self.message}")


class RpcTimeoutError(RpcError):
    def __init__(self, method_name):
        super().__init__(method_name, "RPC_TIMEOUT", "The RPC method timed out.")


class RpcClientError(RpcError):
    pass


class RpcServerError(RpcError):
    pass


class RpcProxy:
    def __init__(self, messenger, service_name=None):
        self.messenger = messenger
        self.service_name = service_name

    def bind(self, service_name):
        self.service_name = service_name
        return self

    def __getattr__(self, method_name):
        if self.service_name is None:
            raise AttributeError(
                f"Service name not set for {self.messenger.service_name}"
            )

        async def _call(*args, extra_headers=None, **kwargs):
            kwargs.pop("extra_headers", None)
            return await self.messenger.call(
                self.service_name,
                method_name,
                *args,
                extra_headers=extra_headers,
                **kwargs,
            )

        return _call



import asyncio
import signal

from kontiki.container import ServiceContainer
from kontiki.utils import log, setup_logger

# -----------------------------------------------------------------------------


async def run_service(
    service_cls, config_paths, version, disable_service_registration=False
):
    setup_logger()
    container = ServiceContainer(
        service_cls, version, config_paths, disable_service_registration
    )
    shutdown_event = asyncio.Event()

    def setup_signal_handlers():
        # Sets up signal handlers for clean shutdown.
        # SIGTERM: Typically sent by the system when stopping the service
        # (e.g., via `kill`).
        # SIGINT: Triggered when the user presses Ctrl+C in the terminal.
        # These signals will set the `shutdown_event`, allowing the service
        # to terminate properly.

        loop = asyncio.get_running_loop()
        loop.add_signal_handler(signal.SIGTERM, shutdown_event.set)
        loop.add_signal_handler(signal.SIGINT, shutdown_event.set)

    setup_signal_handlers()

    try:
        await container.setup()
    except Exception as e:
        # Log with full traceback, then let the exception bubble up so the CLI
        # entrypoint can display a clear error message.
        log.error("Exception occurred during service setup: %s", e, exc_info=True)
        await container.stop()
        raise

    await container.start()

    try:
        await shutdown_event.wait()
    finally:
        log.info("Service shutting down")
        await container.stop()



import inspect
import logging.config
import os
import socket
import uuid

from kontiki.configuration.configuration import DEFAULT_LOGGING_CONFIGURATION
from kontiki.configuration.merge import merge
from kontiki.configuration.parameter import get_kontiki_parameter
from kontiki.delegate import ServiceDelegate
from kontiki.messaging.consumer.core import Consumer
from kontiki.registry.client.heartbeat_publisher import HeartbeatPublisher
from kontiki.registry.client.registry_client import ServiceRegistryClient
from kontiki.task.task import Task
from kontiki.utils import log
from kontiki.web.web import HttpServer

# -----------------------------------------------------------------------------


class ServiceContainer:
    def __init__(
        self,
        service_cls,
        version,
        config_paths,
        disable_service_registration,
        config=None,
    ):
        self.service_cls = service_cls
        self.service_instance = service_cls()
        self.service_name = (
            service_cls.name
            if hasattr(service_cls, "name")
            else self.service_instance.__class__.__name__
        )
        self.http_server = None
        # Unique ID for each instance of the service
        self.instance_id = str(uuid.uuid4())
        self.pid = os.getpid()
        self.host = socket.gethostname()
        self.version = version

        self.config = {}
        self.delegates = {}
        self.tasks = []
        self.disable_service_registration = disable_service_registration

        self.service_registry_client = None
        if config_paths:
            self.config = self.load_config_files(config_paths)
        elif config:
            self.config = config
        else:
            log.error("No service configuration provided.")
            raise RuntimeError("No service configuration provided.")

        # Initialize logging system
        logging_config = self.config.get("logging", DEFAULT_LOGGING_CONFIGURATION)
        logging.config.dictConfig(logging_config)

    def load_config_files(self, conf_files):
        try:
            conf = merge(conf_files)
            log.info("Configurations loaded from: %s", conf_files)
            return conf
        except RuntimeError as e:
            log.error("Error loading configurations: %s", e)
            raise e

    # --------------------------------------------------------------------------
    # Setup
    # --------------------------------------------------------------------------

    async def setup(self):
        # Inject configuration into the service
        if hasattr(self.service_instance, "config"):
            self.service_instance.config = self.config

        self.amqp_consumer = Consumer(self)

        await self.setup_service_registry()
        await self.setup_http_endpoints()
        await self.setup_amqp_endpoints()
        await self.setup_delegates()

        log.info("Service setup completed")

    async def setup_http_endpoints(self):
        if self.has_endpoints("http"):
            http_endpoints = self.get_endpoints("http")
            self.http_server = HttpServer(self, http_endpoints)
            await self.http_server.setup()

    async def setup_amqp_endpoints(self):
        if self.has_endpoints("on_event") or self.has_endpoints("rpc"):
            # Setup AMQP consumer
            await self.amqp_consumer.setup()

            # Setup on_event endpoints
            on_event_tasks = self.get_endpoints("on_event")
            await self.amqp_consumer.add_on_event_tasks(on_event_tasks)

            # Setup rpc endpoints
            remote_tasks = self.get_endpoints("rpc")
            await self.amqp_consumer.add_rpc_tasks(remote_tasks)

    async def setup_delegates(self):
        # Bind delegates
        for attr_name, delegate in self.get_delegates().items():
            delegate.bind(self, attr_name)
            log.debug("Binding %s and %s", self, attr_name)
            self.delegates[attr_name] = delegate
            setattr(self, attr_name, delegate)

        # Setup delegates
        for delegate in self.delegates.values():
            await delegate.setup()

    async def setup_service_registry(self):
        self.disable_service_registration = get_kontiki_parameter(
            self.config, "registration.disable", self.disable_service_registration
        )
        if self.disable_service_registration:
            return
        self.service_registry_client = ServiceRegistryClient(self)
        await self.service_registry_client.setup()

    # --------------------------------------------------------------------------
    # Start
    # --------------------------------------------------------------------------

    async def start(self):
        if self.http_server is not None:
            await self.http_server.start()
        if self.amqp_consumer is not None:
            await self.amqp_consumer.start()

        await self.start_tasks()

        for delegate in self.delegates.values():
            await delegate.start()

    async def start_tasks(self):
        for attr_name in dir(self.service_instance):
            attr = getattr(self.service_instance, attr_name)
            if hasattr(attr, "_task_interval"):
                interval = getattr(attr, "_task_interval")
                immediate = getattr(attr, "_task_immediate", True)

                task = Task(interval, attr, immediate)
                log.debug("Starting %s task.", attr)
                self.tasks.append(task)
                task.start()

    # --------------------------------------------------------------------------
    # Stop
    # --------------------------------------------------------------------------

    async def stop(self):
        log.info("Stopping the service...")
        for delegate in self.delegates.values():
            await delegate.stop()

        for task in self.tasks:
            task.stop()

        if self.http_server:
            await self.http_server.stop()

        if self.service_registry_client:
            await self.service_registry_client.unregister()
            await self.service_registry_client.stop()
        if self.amqp_consumer:
            await self.amqp_consumer.stop()

    # Internal methods

    def get_endpoints(self, type_):
        endpoints = []
        for name, method in inspect.getmembers(
            self.service_cls, predicate=inspect.isfunction
        ):
            if hasattr(method, f"_{type_}_endpoint"):
                log.debug("Discovered %s entrypoint: %s", type_, name)
                endpoints.append(method)
        return endpoints

    def has_endpoints(self, type_):
        for _, method in inspect.getmembers(
            self.service_cls, predicate=inspect.isfunction
        ):
            if hasattr(method, f"_{type_}_endpoint"):
                return True
        return False

    def get_delegates(self):
        delegates = self._get_internal_delegates()
        for attr_name, value in inspect.getmembers(self.service_cls):
            if isinstance(value, ServiceDelegate):
                delegates[attr_name] = value
        return delegates

    def _get_internal_delegates(self):
        delegates = {}
        delegates["_kontiki_heartbeat"] = HeartbeatPublisher(
            self.service_registry_client
        )
        return delegates



import asyncio

from aio_pika import IncomingMessage

from kontiki.utils import log


def on_event(
    event_type_or_key,
    use_config: bool = False,
    *,
    include_headers: bool = False,
    requeue_on_error: bool = False,
    reject_on_redelivered: bool = False,
    in_session: bool = False,
    broadcast: bool = False,
):
    """Declare an event handler for a Kontiki service.

    Args:
        event_type_or_key: Static event name or configuration key.
        use_config: When True, resolve event_type_or_key from configuration.
        include_headers: When True, pass AMQP headers to the handler via _headers.
        requeue_on_error: When True, requeue messages on handler error.
        reject_on_redelivered: When True, reject messages that were already redelivered.
        in_session: When True, the event is scoped to a specific service instance
            and session. This is a higher-level flag; internally it implies
            routing to a single instance.
        broadcast: When True, every instance of the service will receive the
            event (no competing consumers within the service).
    """

    def decorator(handler):
        if in_session and broadcast:
            raise ValueError(
                "on_event(): 'in_session' and 'broadcast' cannot both be True."
            )
        handler._on_event_endpoint = {
            "event_type_or_key": event_type_or_key,
            "use_config": use_config,
            # target_instance is an internal detail; it is derived from in_session.
            "target_instance": in_session,
            "include_headers": include_headers,
            "requeue_on_error": requeue_on_error,
            "reject_on_redelivered": reject_on_redelivered,
            "broadcast": broadcast,
        }
        return handler

    return decorator


class OnEventTask:
    def __init__(
        self,
        event_type,
        task,
        queue,
        serializer,
        include_headers,
        requeue_on_error,
        reject_on_redelivered,
    ):
        self.event_type = event_type
        self.task = task
        self.queue = queue
        self.serializer = serializer
        self.include_headers = include_headers
        self.requeue_on_error = requeue_on_error
        self.reject_on_redelivered = reject_on_redelivered

    async def run(self):
        async def consume_message(message: IncomingMessage):
            try:
                log.debug(
                    "Consuming event %s (redelivered=%s, headers=%s)",
                    self.event_type,
                    message.redelivered,
                    message.headers,
                )
                async with message.process(
                    requeue=self.requeue_on_error,
                    reject_on_redelivered=self.reject_on_redelivered,
                ):
                    obj = self.serializer.loads(message.body)
                    log.info(
                        "Message received on %s: %s", self.event_type, message.body
                    )

                    headers = message.headers if self.include_headers else {}
                    if asyncio.iscoroutinefunction(self.task):
                        if headers:
                            await self.task(obj, _headers=headers)
                        else:
                            await self.task(obj)
                    else:
                        if headers:
                            self.task(obj, _headers=headers)
                        else:
                            self.task(obj)

            except Exception as e:
                log.error("Error occurred while consuming the event: %s", e)
                await message.nack(requeue=False)

        await self.queue.consume(consume_message)



import argparse
import asyncio

from kontiki.runner.runner import run_service

# -----------------------------------------------------------------------------


def run(service_cls, description, version, disable_service_registration=False):
    parser = argparse.ArgumentParser(description=description)
    help_ = "Path to a config file. Specify multiple times for multiple files."
    parser.add_argument("--config", action="append", required=True, help=help_)
    args = parser.parse_args()

    asyncio.run(
        run_service(service_cls, args.config, version, disable_service_registration)
    )



import logging

KONTIKI = "kontiki"

log = logging.getLogger(KONTIKI)


def setup_logger():
    if not log.handlers:
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        log.addHandler(handler)
        log.setLevel(logging.INFO)


def get_kontiki_prefix():
    return f"{KONTIKI}_"


def get_kontiki_header_name(name):
    return f"{get_kontiki_prefix()}{name}"



import asyncio
import logging
from datetime import datetime, timedelta, timezone

from kontiki.configuration.parameter import get_parameter
from kontiki.registry.common import EXCEPTION_RKEY


class ExceptionTracker:
    def __init__(self, core):
        self.core = core
        self.exceptions = []

    async def setup(self):
        self.exception_ttl_minutes = get_parameter(
            self.core.container.config, "exception_tracker.ttl_minutes", 0
        )

        self.exception_ttl_hours = get_parameter(
            self.core.container.config, "exception_tracker.ttl_hours", 24 * 7
        )

        # Use minutes if set, otherwise use hours.
        if self.exception_ttl_minutes > 0:
            self.exception_ttl = self.exception_ttl_minutes
        elif self.exception_ttl_hours > 0:
            self.exception_ttl = self.exception_ttl_hours * 60

        self.disable_tracking = get_parameter(
            self.core.container.config, "exception_tracker.disable", False
        )

        self.cleanup_interval = get_parameter(
            self.core.container.config,
            "exception_tracker.cleanup_interval_seconds",
            3600,
        )

        if self.is_disabled():
            return

        await self.core.create_and_consume_queue(EXCEPTION_RKEY, self._handle_exception)
        self.cleanup_task = asyncio.create_task(self._cleanup_exceptions())
        logging.debug("ExceptionTracker setup completed.")

    async def _handle_exception(self, message):
        async with message.process():
            logging.debug("Received exception: %s", message.body)
            try:
                data = self.core.serializer.loads(message.body)
                await self.store_exception(data)

            except Exception as e:
                logging.error("Error processing exception: %s", e)

    async def store_exception(self, exception_data):
        logging.debug("Storing exception %s", exception_data)
        if self.is_disabled():
            return

        try:
            self.exceptions.append(exception_data)
            logging.debug("Exception recorded: %s", exception_data)
        except Exception as e:
            logging.error("Error recording exception: %s", e)

    async def _cleanup_exceptions(self):
        logging.info(
            "Starting cleanup task with %s seconds." "interval (TTL: %s minutes)",
            self.cleanup_interval,
            self.exception_ttl,
        )

        while True:
            try:
                self._purge_expired_exceptions()
                await asyncio.sleep(self.cleanup_interval)

            except asyncio.CancelledError:
                logging.info("Cleanup task cancelled.")
                break

            except Exception as e:
                logging.error("Error during cleanup task: %s", e)
                # Avoid a tight error loop that can spam logs and fill disk.
                await asyncio.sleep(self.cleanup_interval)

    def is_disabled(self):
        if self.disable_tracking:
            logging.info("Exception tracking is disabled.")
        return self.disable_tracking

    def get_exceptions(self):
        return self.exceptions

    def _purge_expired_exceptions(self):
        expiration_time = datetime.now(timezone.utc) - timedelta(
            minutes=self.exception_ttl
        )

        index = len(self.exceptions)
        cutoff = 0
        for index, exception in enumerate(self.exceptions):
            exception_timestamp = exception.get("timestamp")
            if exception_timestamp is None:
                cutoff = index + 1
                continue

            if isinstance(exception_timestamp, str):
                try:
                    exception_timestamp = datetime.fromisoformat(exception_timestamp)
                except Exception:
                    logging.warning("Invalid timestamp format: %s", exception_timestamp)
                    cutoff = index + 1
                    continue

            # Normalise to UTC-aware datetime to avoid naive/aware comparisons.
            if isinstance(exception_timestamp, datetime):
                if exception_timestamp.tzinfo is None:
                    exception_timestamp = exception_timestamp.replace(
                        tzinfo=timezone.utc
                    )
                else:
                    exception_timestamp = exception_timestamp.astimezone(timezone.utc)

            if exception_timestamp <= expiration_time:
                cutoff = index + 1
                continue

            # Keep exceptions that are not expired. Break the loop.
            break

        # Remove expired exceptions.
        if cutoff > 0:
            del self.exceptions[:cutoff]
            logging.debug("Cleaned up %s expired exceptions.", cutoff)
        else:
            logging.debug("No expired exceptions found.")



import asyncio
import socket
import uuid
from datetime import datetime, timezone

from aio_pika import Message, connect_robust
from aio_pika.exceptions import ChannelInvalidStateError

from kontiki.delegate import ServiceDelegate
from kontiki.messaging.common import (
    AMQP_DEFAULT_URL,
    EVENT_EXCHANGE,
    KONTIKI_SESSION_OPEN_RPC,
    create_tls_context,
    declare_event_exchange,
    declare_rpc_exchange,
    get_amqp_url,
    get_rpc_timeout,
)
from kontiki.messaging.publisher.rpc import (
    RpcClientError,
    RpcServerError,
    RpcTimeoutError,
)
from kontiki.messaging.publisher.session import EventSession
from kontiki.messaging.rpc import RpcErrorType, RpcReturn
from kontiki.messaging.serialization import DEFAULT_SERIALIZATION, Serializer
from kontiki.utils import KONTIKI, get_kontiki_header_name, log, setup_logger


class Messenger(ServiceDelegate):
    def __init__(
        self,
        amqp_url=AMQP_DEFAULT_URL,
        event_exchange=EVENT_EXCHANGE,
        serialization=DEFAULT_SERIALIZATION,
        standalone=False,
        client_name="client",
    ):
        super().__init__()
        self.connection = None
        self.channel = None
        self.futures = {}
        self.callback_queue = None
        self.event_exchange_name = event_exchange
        self.serializer = None
        self.serialization = serialization
        self.amqp_url = amqp_url
        self.standalone = standalone
        self._reconnecting = False
        self._started = False

        # Identification differs between standalone and container-attached modes.
        if self.standalone:
            base_name = client_name
            if not base_name.startswith(f"{KONTIKI}-standalone-"):
                base_name = f"{KONTIKI}-standalone-{base_name}"
            self.service_name = base_name
            # Give each standalone client a stable instance id if not provided
            self.instance_id = str(uuid.uuid4())

    async def _setup(self, config):
        amqp_url = get_amqp_url(config, self.amqp_url)

        # Create SSL context if configured.
        tls_ctx = create_tls_context(config)

        self.connection = await connect_robust(amqp_url, ssl_context=tls_ctx)
        self.channel = await self.connection.channel()
        # Create exchanges for async and sync communications
        self.event_exchange = await declare_event_exchange(
            self.channel, self.event_exchange_name
        )
        self.rpc_exchange = await declare_rpc_exchange(self.channel)
        # Declare shared callback queue
        self.callback_queue = await self.channel.declare_queue(
            exclusive=True, auto_delete=True
        )
        # Sets serializer
        self.serializer = Serializer(config, serialization=self.serialization)
        # Cache RPC timeout from config (works both in container and standalone)
        self._rpc_timeout = get_rpc_timeout(config)

    async def setup(self):
        if self._started:
            return
        if self.standalone:
            if not log.hasHandlers():
                setup_logger()
            await self._setup({KONTIKI: {"amqp": {"url": self.amqp_url}}})
        else:
            await self._setup(self.container.config)
        self._started = True

    async def start(self):
        # Alias for standalone clients: start/stop feels more natural than setup/stop.
        await self.setup()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()
        return False

    async def stop(self):
        if self.connection:
            await self.connection.close()
            log.debug("AMQP Event Dispatcher connection closed.")
        self.connection = None
        self.channel = None
        self.callback_queue = None
        self._started = False

    async def publish(self, event_type, obj, reply_to=None, extra_headers=None):
        if extra_headers is None:
            extra_headers = {}

        event_headers = {
            "event_type": event_type,
            # reply_to is kept for advanced/legacy AMQP usage.
            # Normal Kontiki flows should prefer call() for request/reply.
            "reply_to": reply_to,
        }
        headers = self.get_service_headers() | event_headers | extra_headers
        message = self.serializer.dumps(obj)

        try:
            await self.event_exchange.publish(
                Message(body=message, headers=headers), routing_key=event_type
            )
        except ChannelInvalidStateError:
            log.info("Channel is in an invalid state. Attempting to reconnect...")
            await self.reconnect()
            await self.event_exchange.publish(
                Message(body=message, headers=headers), routing_key=event_type
            )

        log.debug("Event published: %s -> %s", event_type, message)

    async def call(
        self, service_name, method_name, *args, extra_headers=None, **kwargs
    ):
        cid = str(uuid.uuid4())
        loop = asyncio.get_running_loop()

        # Create a future to wait for the response
        future = loop.create_future()
        self.futures[cid] = future

        async def on_response(message):
            cid = message.correlation_id
            if cid in self.futures:
                future = self.futures.pop(cid)
                if not future.done():
                    try:
                        response = self.serializer.loads(message.body)
                        future.set_result(response)
                        log.debug(
                            "Response received for correlation_id=%s: %s", cid, response
                        )
                    except Exception as e:
                        future.set_exception(e)
                        log.error("Error processing response: %s", e)
                else:
                    log.warning("Response for %s already handled.", cid)
            else:
                log.warning("Unknown correlation_id: %s", cid)

        # Start consuming messages from the temporary queue
        await self.callback_queue.consume(on_response)

        # Sends the sync request
        if extra_headers is None:
            extra_headers = {}
        remote_headers = {"remote_method": method_name}
        headers = self.get_service_headers() | remote_headers | extra_headers

        routing_key = f"{service_name}.{method_name}"
        request_body = self.serializer.dumps({"args": args, "kwargs": kwargs})
        request_message = Message(
            body=request_body,
            correlation_id=cid,
            reply_to=self.callback_queue.name,
            headers=headers,
        )

        log.debug("Call: %s(args=%s, kwargs=%s)", method_name, args, kwargs)
        await self.rpc_exchange.publish(request_message, routing_key=routing_key)

        try:
            # Wait for the response with a timeout
            response = await asyncio.wait_for(future, timeout=self._rpc_timeout)
        except asyncio.TimeoutError:
            # Cleanup future on timeout
            if cid in self.futures:
                del self.futures[cid]
            raise RpcTimeoutError(method_name)

        if isinstance(response, RpcReturn):
            if response.success:
                return response.result

            if response.error_type == RpcErrorType.CLIENT:
                raise RpcClientError(method_name, response.error_code, response.message)
            else:
                raise RpcServerError(method_name, response.error_code, response.message)

        return response

    async def open_session(self, service_name):
        # Open an event session with a given service.
        # Calls the internal Kontiki RPC to open a session.
        instance_id, session_id = await self.call(
            service_name, KONTIKI_SESSION_OPEN_RPC
        )
        return EventSession(self, service_name, instance_id, session_id)

    async def reconnect(self):
        if self._reconnecting:
            return
        self._reconnecting = True
        try:
            log.info("Reconnecting to AMQP server...")
            if self.connection:
                await self.connection.close()
            await self.setup()
            log.info("Reconnected successfully.")
        finally:
            self._reconnecting = False

    def get_service_headers(self):
        now_utc_iso = datetime.now(timezone.utc).isoformat()
        if self.container:
            return {
                get_kontiki_header_name("service_name"): self.container.service_name,
                get_kontiki_header_name("instance_id"): str(self.container.instance_id),
                get_kontiki_header_name("host"): self.container.host,
                get_kontiki_header_name("timestamp"): now_utc_iso,
            }
        return {
            get_kontiki_header_name("service_name"): self.service_name,
            get_kontiki_header_name("instance_id"): self.instance_id,
            get_kontiki_header_name("host"): socket.gethostname(),
            get_kontiki_header_name("timestamp"): now_utc_iso,
        }



from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Optional


class RpcErrorType(Enum):
    NONE = auto()
    CLIENT = auto()
    SERVER = auto()


@dataclass
class RpcReturn:
    success: bool
    result: Optional[Any] = None
    message: Optional[str] = None
    error_type: RpcErrorType = RpcErrorType.NONE
    error_code: Optional[str] = None



import asyncio

from kontiki.delegate import ServiceDelegate
from kontiki.registry.common import get_heartbeat_interval
from kontiki.utils import log

# -----------------------------------------------------------------------------


def degraded_on(func):
    func._degraded = True
    return func


# -----------------------------------------------------------------------------


class HeartbeatPublisher(ServiceDelegate):
    def __init__(self, service_registry_client):
        self.interval = None
        self.task = None
        self.service_registry_client = service_registry_client
        self.stop_condition = None
        self._is_degraded = False
        super().__init__()

    async def setup(self):
        if self.service_registry_client:
            self.interval = get_heartbeat_interval(self.container.config)
            log.debug("Heartbeat interval set to %s", self.interval)

            service_instance = self.container.service_instance
            for attr_name in dir(service_instance):
                attr = getattr(service_instance, attr_name)
                if callable(attr) and getattr(attr, "_degraded", False):
                    self.stop_condition = attr
                    break
        else:
            log.info("Service registration is disabled.")

    async def start(self):
        if self.service_registry_client:
            self.task = asyncio.create_task(self._send_heartbeat())

    async def stop(self):
        if self.task:
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError:
                pass

    async def _send_heartbeat(self):
        log.info("Starting heartbeat task.")
        error_logged = False
        while True:
            try:
                should_degrade = not self._should_send_heartbeat()

                # Normal state -> degraded state
                if should_degrade and not self._is_degraded:
                    self._is_degraded = True
                    log.info("Service entered degraded state.")

                # Degraded state -> Normal state
                elif not should_degrade and self._is_degraded:
                    self._is_degraded = False
                    log.info("Service recovered from degraded state.")

                await self.service_registry_client.heartbeat(self._is_degraded)

                if self.interval is not None:
                    await asyncio.sleep(self.interval)
                else:
                    # Default to 10 seconds if interval is not set
                    await asyncio.sleep(10)

            except asyncio.CancelledError:
                log.info("Heartbeat task was cancelled.")
                break
            except Exception as e:
                # To avoid log messages to be printed indefinitely.
                if not error_logged:
                    log.error("Error in heartbeat task: %s", e)
                    error_logged = True

    def _should_send_heartbeat(self):
        if not self.stop_condition:
            return True
        return not self.stop_condition()



from kontiki.testing.delegates.event_manager import EventManager
from kontiki.testing.delegates.http_manager import HttpManager
from kontiki.testing.delegates.rpc_manager import RpcManager


class MockService:
    event_manager = EventManager()
    remote_call_manager = RpcManager()
    http_manager = HttpManager()

    def get_events(self, wait_for_events=0, timeout=None):
        return self.event_manager.get_events(wait_for_events, timeout)

    def clean_events(self):
        self.event_manager.clean()

    def add_http_response(self, response):
        self.http_manager.add_response(response)

    def get_http_requests(self):
        return self.http_manager.get_requests()

    def clean_http_requests(self):
        self.http_manager.clean()

    def add_remote_return_value(self, value):
        self.remote_call_manager.add_return_value(value)

    def store_call_args(self, *args, **kwargs):
        self.remote_call_manager.store_call_args(args, kwargs)

    def get_return_value(self):
        return self.remote_call_manager.get_return_value()

    def get_remote_calls(self):
        return self.remote_call_manager.get_call_args()

    def clean_remote_calls(self):
        self.remote_call_manager.clean()



import asyncio
from datetime import datetime

from aio_pika import Message, connect_robust

from kontiki.configuration.parameter import get_kontiki_parameter
from kontiki.messaging.common import create_tls_context, get_amqp_url
from kontiki.messaging.serialization import Serializer
from kontiki.registry.common import (
    EXCEPTION_RKEY,
    HEARTBEAT_RKEY,
    REGISTER_RKEY,
    UNREGISTER_RKEY,
    declare_registry_admin_exchange,
    get_heartbeat_interval,
)
from kontiki.utils import log

# -----------------------------------------------------------------------------


def filter_config_by_whitelist(config, whitelist_patterns):
    if not whitelist_patterns:
        return {}

    def _is_included(path):
        for pattern in whitelist_patterns:
            if path == pattern or path.startswith(f"{pattern}."):
                return True
        return False

    def _filter_dict(d, parent_path=""):
        filtered = {}
        for key, value in d.items():
            current_path = f"{parent_path}.{key}" if parent_path else key

            if _is_included(current_path):
                if isinstance(value, dict):
                    filtered[key] = _filter_dict(value, current_path)
                else:
                    filtered[key] = value
            elif isinstance(value, dict):
                nested_filtered = _filter_dict(value, current_path)
                if nested_filtered:
                    filtered[key] = nested_filtered
        return filtered

    return _filter_dict(config)


def publish(routing_key):
    def decorator(func):
        async def wrapper(self, *args, **kwargs):
            if not self.registry_admin_exchange:
                msg = "Registry Admin Exchange not setup."
                log.error(msg)
                raise RuntimeError(msg)

            body = await func(self, *args, **kwargs)

            message = Message(body=self.serializer.dumps(body))
            await self.registry_admin_exchange.publish(message, routing_key=routing_key)
            log.info("Published message to %s: %s", routing_key, body)

        return wrapper

    return decorator


# -----------------------------------------------------------------------------


class ServiceRegistryClient:
    def __init__(self, container):
        self.container = container
        self.connection = None
        self.channel = None
        self.registry_admin_exchange = None

    async def setup(self):
        log.debug("ServiceRegistryClient Setup")
        amqp_url = get_amqp_url(self.container.config)
        ssl_ctx = create_tls_context(self.container.config)
        self.connection = await connect_robust(amqp_url, ssl_context=ssl_ctx)
        self.channel = await self.connection.channel()
        await self.channel.set_qos(prefetch_count=10)
        self.registry_admin_exchange = await declare_registry_admin_exchange(
            self.channel
        )
        self.serializer = Serializer(self.container.config)

        queue_name = "register_again.queue"
        queue = await self.channel.declare_queue(queue_name, durable=True)
        service_name = self.container.service_name
        instance_id = self.container.instance_id
        routing_key = f"{service_name}.{instance_id}.register_again"
        await queue.bind(self.registry_admin_exchange, routing_key=routing_key)
        await queue.consume(self._register_again)

        # Delay registration to ensure the ServiceRegistry is fully ready to
        # receive messages. This helps avoid lost registration messages during startup.

        delay = get_kontiki_parameter(
            self.container.config, "registration.delay", default=2
        )
        log.info("Delaying registration by %s seconds.", delay)
        await asyncio.sleep(delay)
        await self.register()

    async def stop(self):
        if self.connection:
            await self.connection.close()

    def _get_config(self):
        config = self.container.config
        public_paths = get_kontiki_parameter(
            config, "registration.configuration.public_paths", default=None
        )
        if public_paths:
            return filter_config_by_whitelist(config, public_paths)
        return {}

    @publish(REGISTER_RKEY)
    async def register(self):
        heartbeat_interval = get_heartbeat_interval(self.container.config)
        filtered_config = self._get_config()
        body = {
            "service_name": self.container.service_name,
            "instance_id": self.container.instance_id,
            "pid": self.container.pid,
            "host": self.container.host,
            "service_version": self.container.version,
            "heartbeat_interval": heartbeat_interval,
        }

        if filtered_config:
            body["config"] = filtered_config
        return body

    @publish(UNREGISTER_RKEY)
    async def unregister(self):
        body = {
            "service_name": self.container.service_name,
            "instance_id": self.container.instance_id,
        }
        return body

    @publish(HEARTBEAT_RKEY)
    async def heartbeat(self, degraded):
        body = {
            "service_name": self.container.service_name,
            "instance_id": self.container.instance_id,
            "degraded": degraded,
        }
        return body

    @publish(EXCEPTION_RKEY)
    async def register_exception(self, exception, context):
        body = {
            "service_name": self.container.service_name,
            "instance_id": self.container.instance_id,
            "exception_type": type(exception).__name__,
            "message": str(exception),
            "context": context,
            "timestamp": datetime.now().isoformat(),
        }
        return body

    async def _register_again(self, message):
        async with message.process():
            try:
                service_name = self.container.service_name
                instance_id = self.container.instance_id
                service_str = f"{service_name}#{instance_id}"
                log.info("Re-registering service %s.", service_str)
                await self.register()
            except Exception as e:
                log.error("Error handling register_again message: %s", e)



import asyncio

from kontiki.utils import log

# -----------------------------------------------------------------------------


class Task:
    def __init__(self, interval, user_task, immediate=True):
        self.interval = interval
        self.user_task = user_task
        self.immediate = immediate
        self.running = False
        self.timer_loop_task = None

    def start(self):
        if self.running:
            log.error("Repeat task already running")
            return

        self.running = True
        self.timer_loop_task = asyncio.create_task(self._run())

    def stop(self):
        self.running = False
        if self.timer_loop_task:
            self.timer_loop_task.cancel()
            self.timer_loop_task = None

    async def _run(self):
        if self.immediate:
            await self._execute_user_task()

        while self.running:
            await asyncio.sleep(self.interval)
            await self._execute_user_task()

    async def _execute_user_task(self):
        try:
            if asyncio.iscoroutinefunction(self.user_task):
                await self.user_task()
            else:
                self.user_task()
        except Exception as e:
            log.error("Repeat task: Error executing user task: %s", e)


def task(interval, immediate=True):
    # Mark a method as a timer. Timer will be instanciated regarding those
    # informations
    def decorator(func):
        func._task_interval = interval
        func._task_immediate = immediate
        return func

    return decorator



from kontiki.messaging.consumer.event import on_event
from kontiki.messaging.consumer.rpc import rpc, rpc_error
from kontiki.messaging.publisher.messenger import Messenger
from kontiki.messaging.publisher.rpc import (
    RpcClientError,
    RpcProxy,
    RpcServerError,
    RpcTimeoutError,
)
from kontiki.messaging.publisher.session import EventSession

__all__ = [
    "EventSession",
    "Messenger",
    "RpcClientError",
    "RpcProxy",
    "RpcServerError",
    "RpcTimeoutError",
    "on_event",
    "rpc",
    "rpc_error",
]



# Re-exports for public API.
from .manager import MockServiceManager  # noqa: F401
from .mock import MockService  # noqa: F401
from .runner import MockServiceRunner  # noqa: F401

__all__ = ["MockServiceManager", "MockService", "MockServiceRunner"]



from kontiki.messaging.publisher.rpc import RpcProxy

# -----------------------------------------------------------------------------


class ServiceRegistryProxy(RpcProxy):
    def __init__(self, messenger):
        super().__init__(messenger, service_name="ServiceRegistry")



import asyncio

from aio_pika import Message

from kontiki.messaging.rpc import RpcErrorType, RpcReturn
from kontiki.utils import log


def rpc(_func=None, *, include_headers=False):
    def decorator(handler):
        handler._rpc_endpoint = {
            "name": handler.__name__,
            "include_headers": include_headers,
        }
        return handler

    if _func is None:
        return decorator
    return decorator(_func)


def rpc_error(error_code: str, msg: str):
    return RpcReturn(
        success=False,
        message=msg,
        error_type=RpcErrorType.CLIENT,
        error_code=error_code,
    )


class RpcTask:
    def __init__(
        self, name, task, container, reply_exchange, queue, serializer, include_headers
    ):
        self.container = container
        self.task = task
        self.reply_exchange = reply_exchange
        self.queue = queue
        self.futures = {}
        self.name = name
        self.serializer = serializer
        self.include_headers = include_headers

    async def run(self):
        async def handle_rpc(message):
            async with message.process():
                # Identify each request/response pair.
                cid = message.correlation_id
                log.debug("Message received for correlation_id=%s", cid)
                if cid in self.futures:
                    log.error("Duplicate message (%s)", cid)
                    return

                loop = asyncio.get_running_loop()
                future = loop.create_future()
                self.futures[cid] = future

                # Parse the request
                try:
                    request = self.serializer.loads(message.body)
                    args = request.get("args", [])
                    kwargs = request.get("kwargs", {})
                    headers = message.headers if self.include_headers else None
                except Exception as e:
                    log.error("Invalid RPC message format: %s", e)
                    del self.futures[cid]
                    return

                msg = "RPC request received: method=%s args=%s, kwargs=%s"
                log.info(msg, self.name, args, kwargs)

                # Call the handler
                try:
                    if headers:
                        response_data = await self.task(
                            self.container, *args, **kwargs, _headers=headers
                        )
                    else:
                        response_data = await self.task(self.container, *args, **kwargs)

                    if not isinstance(response_data, RpcReturn):
                        response_data = RpcReturn(success=True, result=response_data)

                    if response_data is None:
                        response_data = ""
                    if not future.done():
                        future.set_result(response_data)

                except Exception as e:
                    msg = "Error while processing RPC method %s: %s"
                    log.error(msg, self.name, e)
                    if not future.done():
                        future.set_result(
                            RpcReturn(
                                success=False,
                                message=str(e),
                                error_type=RpcErrorType.SERVER,
                                error_code="INTERNAL_ERROR",
                            )
                        )

                # Send the response
                if message.reply_to:
                    try:
                        response_data = await future
                        serialized = self.serializer.dumps(response_data)
                        log.debug("Reply %s to %s", serialized, message.reply_to)
                        response_message = Message(body=serialized, correlation_id=cid)
                        await self.reply_exchange.publish(
                            response_message, routing_key=message.reply_to
                        )
                    except Exception as e:
                        log.error("Failed to send response: %s", e)
                    finally:
                        if cid in self.futures:
                            del self.futures[cid]

        # Start consuming messages for RPC
        await self.queue.consume(handle_rpc)



import asyncio
import logging
import threading
from concurrent.futures import Future as ConcurrentFuture

from kontiki.testing.manager import MockServiceManager

# -----------------------------------------------------------------------------
# Runner for mock service to be used in synchronous environment.
# Initially implemented for behave integration test environment.
# -----------------------------------------------------------------------------


class MockServiceRunner(threading.Thread):
    def __init__(self, mock_manager: MockServiceManager):
        super().__init__(daemon=True)
        self.mock_manager = mock_manager
        self.loop = asyncio.new_event_loop()
        self.result = None
        self.ready_event = threading.Event()
        self.stop_event = threading.Event()

    def publish(self, event_type, obj, reply_to=None, extra_headers=None):
        if extra_headers is None:
            extra_headers = {}
        # Use the runner asynchronous environment to publish input events for testing.
        logging.info("Runner publish call %s - %s", event_type, obj)

        def _publish_task():
            try:
                asyncio.create_task(
                    self.mock_manager.publish(event_type, obj, reply_to, extra_headers)
                )
                logging.info("Publish task successfully scheduled.")
            except Exception as e:
                logging.error("Failed to schedule publish task: %s", e)

        if not self.loop.is_running():
            logging.error("Event loop is not running. Cannot publish.")
            raise RuntimeError("Event loop is not running.")

        self.loop.call_soon_threadsafe(_publish_task)

    def call(self, service_name, method_name, *args, extra_headers=None, **kwargs):
        logging.info(
            "Runner remote call %s.%s with args=%s, kwargs=%s and extra_headers=%s",
            service_name,
            method_name,
            args,
            kwargs,
            extra_headers,
        )

        future = ConcurrentFuture()
        ready_event = threading.Event()

        def _done_callback(fut):
            try:
                future.set_result(fut.result())
            except Exception as e:
                future.set_exception(e)
            finally:
                ready_event.set()

        def _schedule_call():
            task = asyncio.create_task(
                self.mock_manager.call(
                    service_name,
                    method_name,
                    *args,
                    extra_headers=extra_headers,
                    **kwargs,
                )
            )
            task.add_done_callback(_done_callback)

        if not self.loop.is_running():
            logging.error("Event loop is not running. Cannot call remote method.")
            raise RuntimeError("Event loop is not running.")

        self.loop.call_soon_threadsafe(_schedule_call)

        ready_event.wait(timeout=5)
        if not future.done():
            raise TimeoutError("Remote call timed out")
        return future.result()

    # Overridden method that is called when we call start on this object (as a Thread)
    def run(self):
        # Runs the mock services within the thread.
        asyncio.set_event_loop(self.loop)

        try:
            self.result = self.loop.run_until_complete(self.mock_manager.start())
            self.ready_event.set()
            self.loop.run_until_complete(self._wait_for_stop())
            self.loop.run_until_complete(self.mock_manager.stop())
        finally:
            # Stop pending tasks before closing the loop.
            pending_tasks = asyncio.all_tasks(self.loop)
            for task in pending_tasks:
                task.cancel()
            self.loop.run_until_complete(
                asyncio.gather(*pending_tasks, return_exceptions=True)
            )
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())
            self.loop.close()

    async def _wait_for_stop(self):
        while not self.stop_event.is_set():
            await asyncio.sleep(0.1)

    def stop(self):
        self.stop_event.set()
        self.join()
        self._cleanup_async_tasks()

    def _cleanup_async_tasks(self):
        try:
            # Use the runner's own event loop if it's still available
            if hasattr(self, "loop") and not self.loop.is_closed():
                try:
                    # Cancel any remaining tasks in the runner's loop
                    pending_tasks = asyncio.all_tasks(self.loop)
                    for task in pending_tasks:
                        task.cancel()
                    self.loop.run_until_complete(
                        asyncio.wait_for(
                            asyncio.gather(*pending_tasks, return_exceptions=True),
                            timeout=1.0,
                        )
                    )

                except (RuntimeError, asyncio.TimeoutError):
                    # Loop might be closed or tasks might not complete in time
                    pass
                except RuntimeError:
                    # Loop is closed, nothing more we can do
                    pass
            else:
                # Fallback: try to clean up any tasks in the current thread's event loop
                try:
                    loop = asyncio.get_running_loop()
                    pending_tasks = asyncio.all_tasks(loop)
                    for task in pending_tasks:
                        task.cancel()
                except RuntimeError:
                    # No running loop, nothing to clean up
                    pass

        except Exception:
            # Ignore any errors during cleanup to avoid masking other issues
            pass



DEFAULT_LOGGING_CONFIGURATION = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "default": {
            "format": (
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s -"
                " %(filename)s:%(lineno)d"
            )
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "level": "DEBUG",
        }
    },
    "root": {
        "handlers": ["console"],
        "level": "DEBUG",
    },
}
